@@ -5,7 +5,7 @@ package OpenAPI::Modern;
 # ABSTRACT: Validate HTTP requests and responses against an OpenAPI document
 # KEYWORDS: validation evaluation JSON Schema OpenAPI Swagger HTTP request response
 
-our $VERSION = '0.006';
+our $VERSION = '0.007';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -17,13 +17,13 @@ no if "$]" >= 5.033001, feature => 'multidimensional';
 no if "$]" >= 5.033006, feature => 'bareword_filehandles';
 use Carp 'croak';
 use Safe::Isa;
-use Ref::Util qw(is_plain_hashref is_plain_arrayref);
+use Ref::Util qw(is_plain_hashref is_plain_arrayref is_ref);
 use List::Util 1.45 qw(first uniq);
 use Scalar::Util 'looks_like_number';
 use Module::Runtime 'require_module';
 use Feature::Compat::Try;
 use JSON::Schema::Modern 0.527;
-use JSON::Schema::Modern::Utilities qw(jsonp canonical_uri E abort get_type);
+use JSON::Schema::Modern::Utilities qw(jsonp canonical_uri E abort);
 use JSON::Schema::Modern::Document::OpenAPI;
 use MooX::HandlesVia;
 use MooX::TypeTiny 0.002002;
@@ -94,7 +94,7 @@ sub validate_request ($self, $request, $options) {
   my $path_template = $options->{path_template};
   my $captures = $options->{path_captures};
   croak 'missing parameter path_template' if not length $path_template;
-  croak 'missing parameter captures ' if not is_plain_hashref($captures);
+  croak 'missing parameter captures' if not is_plain_hashref($captures);
 
   try {
     my $method = lc $request->method;
@@ -324,6 +324,10 @@ sub _validate_parameter_content ($self, $state, $param_obj, $content_ref) {
 sub _validate_body_content ($self, $state, $content_obj, $message) {
   my $content_type = $message->content_type;
 
+  return E({ %$state, data_path => $state->{data_path} =~ s{body}{header/Content-Type}r, keyword => 'content' },
+      'missing header: Content-Type')
+    if not length $content_type;
+
   # TODO: respect wildcard entries in the openapi document
   return E({ %$state, keyword => 'content' }, 'incorrect Content-Type "%s"', $content_type)
     if not exists $content_obj->{$content_type};
@@ -347,14 +351,15 @@ sub _validate_body_content ($self, $state, $content_obj, $message) {
   my $decoded_content_ref = $message->decoded_content(ref => 1);
 
   # decode the charset
-  my $charset = $message->content_charset;
-  try {
-    $decoded_content_ref =
-      \ Encode::decode($charset, $decoded_content_ref->$*, Encode::FB_CROAK | Encode::LEAVE_SRC);
-  }
-  catch ($e) {
-    return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
-      'could not decode content as %s', $charset);
+  if (my $charset = $message->content_charset) {
+    try {
+      $decoded_content_ref =
+        \ Encode::decode($charset, $decoded_content_ref->$*, Encode::FB_CROAK | Encode::LEAVE_SRC);
+    }
+    catch ($e) {
+      return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
+        'could not decode content as %s', $charset);
+    }
   }
 
   my $media_type_decoder = $self->get_media_type($content_type);
@@ -392,7 +397,9 @@ sub _resolve_ref ($self, $ref, $state) {
   abort({ %$state, keyword => '$ref' }, 'EXCEPTION: unable to find resource %s', $uri)
     if not $schema_info;
 
-  ++$state->{depth};
+  abort($state, 'EXCEPTION: maximum evaluation depth exceeded')
+    if $state->{depth}++ > $self->evaluator->max_traversal_depth;
+
   $state->{initial_schema_uri} = $schema_info->{canonical_uri};
   $state->{traversed_schema_path} = $state->{traversed_schema_path}.$state->{schema_path}.jsonp('/$ref');
   $state->{schema_path} = '';
@@ -405,15 +412,12 @@ sub _evaluate_subschema ($self, $data, $schema, $state) {
   return 1 if is_plain_hashref($schema) ? !keys(%$schema) : $schema; # true schema
 
   # treat numeric-looking data as a string, unless "type" explicitly requests number or integer.
-  my $type = get_type($data);
-  if ((grep $type eq $_, qw(string number)) and looks_like_number($data)) {
-    if (is_plain_hashref($schema) and exists $schema->{type} and not is_plain_arrayref($schema->{type})
-        and grep $schema->{type} eq $_, qw(number integer)) {
-      $data = $data+0;
-    }
-    else {
-      $data = $data.'';
-    }
+  if (is_plain_hashref($schema) and exists $schema->{type} and not is_plain_arrayref($schema->{type})
+      and grep $schema->{type} eq $_, qw(number integer) and looks_like_number($data)) {
+    $data = $data+0;
+  }
+  elsif (defined $data and not is_ref($data)) {
+    $data = $data.'';
   }
 
   # TODO: also handle multi-valued elements like headers and query parameters, when type=array requested
@@ -446,7 +450,7 @@ OpenAPI::Modern - Validate HTTP requests and responses against an OpenAPI docume
 
 =head1 VERSION
 
-version 0.006
+version 0.007
 
 =head1 SYNOPSIS
 
